// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_authentication_bloc/src/core/constants/form_field.dart';
import 'package:wyatt_authentication_bloc/src/core/constants/form_name.dart';
import 'package:wyatt_authentication_bloc/src/data/models/account_wrapper_model.dart';
import 'package:wyatt_authentication_bloc/src/domain/data_sources/local/authentication_cache_data_source.dart';
import 'package:wyatt_authentication_bloc/src/domain/data_sources/remote/authentication_remote_data_source.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/account.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/account_wrapper.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/auth_change_event.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

class AuthenticationRepositoryImpl<T extends Object>
    extends AuthenticationRepository<T> {
  AuthenticationRepositoryImpl({
    required this.authenticationCacheDataSource,
    required this.authenticationRemoteDataSource,
    FormRepository? formRepository,
    // ignore: strict_raw_type
    List<FormInput>? extraSignUpInputs,
    FormInputValidator<String?, ValidationError>? customEmailValidator,
    FormInputValidator<String?, ValidationError>? customPasswordValidator,
    AuthChangeListener<T>? onAuthChange,
    AccountStreamTransformer<T>? accountStreamTransformer,
  })  : _authChangeListener = onAuthChange,
        _accountStreamTransformer = accountStreamTransformer {
    _formRepository = formRepository ?? FormRepositoryImpl();
    _accountStreamTransformer ??= (input) => input
            .where((event) => event is! SignUpAuthChangeEvent)
            .map<FutureOrResult<AccountWrapper<T>>>((event) async {
          if (listener == null) {
            return Ok(AccountWrapperModel<T>(event.account, null));
          }
          // Handle sign in, sign out and refresh
          final dataResult = await listener!.call(this, event);
          return dataResult.map((data) {
            authenticationCacheDataSource.storeData(data);
            return AccountWrapperModel(event.account, data);
          });
        });
    if (formRepository != null) {
      return;
    }
    _formRepository
      ..registerForm(
        WyattFormImpl(
          [
            FormInput(
              AuthFormField.email,
              customEmailValidator ?? const Email.pure(),
            ),
            FormInput(
              AuthFormField.password,
              customPasswordValidator ?? const Password.pure(),
            )
          ],
          name: AuthFormName.signInForm,
        ),
      )
      ..registerForm(
        WyattFormImpl(
          [
            FormInput(
              AuthFormField.email,
              customEmailValidator ?? const Email.pure(),
            ),
            FormInput(
              AuthFormField.password,
              customPasswordValidator ?? const Password.pure(),
            ),
            ...extraSignUpInputs ?? []
          ],
          name: AuthFormName.signUpForm,
        ),
      )
      ..registerForm(
        WyattFormImpl(
          [
            FormInput(
              AuthFormField.email,
              customEmailValidator ?? const Email.pure(),
            ),
          ],
          name: AuthFormName.passwordResetForm,
        ),
      );
  }
  final AuthenticationCacheDataSource<T> authenticationCacheDataSource;
  final AuthenticationRemoteDataSource authenticationRemoteDataSource;

  late FormRepository _formRepository;

  AuthChangeListener<T>? _authChangeListener;
  AccountStreamTransformer<T>? _accountStreamTransformer;

  @override
  FormRepository get formRepository => _formRepository;

  @override
  AuthChangeListener<T>? get listener => _authChangeListener;

  @override
  FutureOrResult<Account> signInWithEmailAndPassword({
    required String email,
    required String password,
  }) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account =
              await authenticationRemoteDataSource.signInWithEmailAndPassword(
            email: email,
            password: password,
          );
          await authenticationCacheDataSource.storeAccount(account);
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<void> signOut() =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationRemoteDataSource.signOut();
          await authenticationCacheDataSource.destroy();
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> signUp({
    required String email,
    required String password,
  }) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account = await authenticationRemoteDataSource.signUp(
            email: email,
            password: password,
          );
          await authenticationCacheDataSource.storeAccount(account);
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<void> destroyCache() =>
      Result.tryCatchAsync<void, AppException, AppException>(
        authenticationCacheDataSource.destroy,
        (error) => error,
      );

  @override
  FutureOrResult<AccountWrapper<T>> getCache() =>
      Result.tryCatchAsync<AccountWrapper<T>, AppException, AppException>(
        authenticationCacheDataSource.load,
        (error) => error,
      );

  @override
  FutureOrResult<Account> getAccount() =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        authenticationCacheDataSource.loadAccount,
        (error) => error,
      );

  @override
  FutureOrResult<void> setAccount(
    Account account,
  ) =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationCacheDataSource.storeAccount(account);
        },
        (error) => error,
      );

  @override
  FutureOrResult<T> getData() =>
      Result.tryCatchAsync<T, AppException, AppException>(
        authenticationCacheDataSource.loadData,
        (error) => error,
      );

  @override
  FutureOrResult<void> setData(
    T? data,
  ) =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationCacheDataSource.storeData(data);
        },
        (error) => error,
      );

  @override
  FutureOrResult<String> getIdentityToken() =>
      Result.tryCatchAsync<String, AppException, AppException>(
        authenticationRemoteDataSource.getIdentityToken,
        (error) => error,
      );

  @override
  Stream<FutureOrResult<AccountWrapper<T>>> streamAccount() =>
      _accountStreamTransformer!.call(changes());

  @override
  FutureOrResult<void> confirmPasswordReset({
    required String code,
    required String newPassword,
  }) =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationRemoteDataSource.confirmPasswordReset(
            code: code,
            newPassword: newPassword,
          );
        },
        (error) => error,
      );

  @override
  FutureOrResult<void> sendEmailVerification() =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationRemoteDataSource.sendEmailVerification();
        },
        (error) => error,
      );

  @override
  FutureOrResult<void> sendPasswordResetEmail({required String email}) =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationRemoteDataSource.sendPasswordResetEmail(
            email: email,
          );
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> signInAnonymously() =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account =
              await authenticationRemoteDataSource.signInAnonymously();
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> signInWithGoogle() =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account =
              await authenticationRemoteDataSource.signInWithGoogle();
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<bool> verifyPasswordResetCode({required String code}) =>
      Result.tryCatchAsync<bool, AppException, AppException>(
        () async {
          final response = await authenticationRemoteDataSource
              .verifyPasswordResetCode(code: code);
          return response;
        },
        (error) => error,
      );

  @override
  FutureOrResult<void> refresh() =>
      Result.tryCatchAsync<void, AppException, AppException>(
        () async {
          await authenticationRemoteDataSource.refresh();
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> reauthenticateWithCredential() =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account = await authenticationRemoteDataSource
              .reauthenticateWithCredential();
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> updateEmail({required String email}) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account =
              await authenticationRemoteDataSource.updateEmail(email: email);
          return account;
        },
        (error) => error,
      );

  @override
  FutureOrResult<Account> updatePassword({required String password}) =>
      Result.tryCatchAsync<Account, AppException, AppException>(
        () async {
          final account = await authenticationRemoteDataSource.updatePassword(
            password: password,
          );
          return account;
        },
        (error) => error,
      );

  Account? _lastChange;

  @override
  Stream<AuthChangeEvent> changes() => authenticationRemoteDataSource
          .streamAccount()
          .map<AuthChangeEvent>((account) {
        if (_lastChange != null && account == null) {
          _lastChange = null;
          return SignOutAuthChangeEvent();
        }
        if (_lastChange == null && account != null) {
          _lastChange = account;
          if (account.isNewUser ?? false) {
            return SignUpAuthChangeEvent(account);
          }
          return SignInAuthChangeEvent(account);
        }
        if (_lastChange != null && account != null) {
          _lastChange = account;
          return RefreshAuthChangeEvent(account);
        }
        if (_lastChange == null && account == null) {
          _lastChange = account;
          return StillUnauthenticatedAuthChangeEvent();
        }
        _lastChange = account;
        return RefreshAuthChangeEvent(account);
      });

  @override
  FutureOrResult<void> addAuthChangeListener(AuthChangeListener<T> listener) {
    _authChangeListener = listener;
    return const Ok(null);
  }

  @override
  FutureOrResult<void> removeAuthChangeListener() {
    _authChangeListener = null;
    return const Ok(null);
  }
}
